// DragDropView.cpp : implementation file
//

#include "stdafx.h"
#include "dragdrop.h"
#include "DragDropView.h"

#ifdef _DEBUG
#define new DEBUG_NEW
#undef THIS_FILE
static char THIS_FILE[] = __FILE__;
#endif

/////////////////////////////////////////////////////////////////////////////
// CDragDropView

IMPLEMENT_DYNCREATE(CDragDropView, CFormView)

CDragDropView::CDragDropView()
	: CFormView(CDragDropView::IDD)
{
	//{{AFX_DATA_INIT(CDragDropView)
	//}}AFX_DATA_INIT
  m_boDragging   = false;
  m_pDragImgList = NULL;
	m_imgList.Create(IDB_SCENE_TREE,16,1,RGB(255,255,255));
}

//

CDragDropView::~CDragDropView()
{
  delete m_pDragImgList;
  m_imgList.DeleteImageList();
}

//

void CDragDropView::DoDataExchange(CDataExchange* pDX)
{
	CFormView::DoDataExchange(pDX);
	//{{AFX_DATA_MAP(CDragDropView)
	DDX_Control(pDX, IDC_TREE, m_Tree);
	//}}AFX_DATA_MAP
}

//

BEGIN_MESSAGE_MAP(CDragDropView, CFormView)
	//{{AFX_MSG_MAP(CDragDropView)
	ON_NOTIFY(TVN_KEYDOWN, IDC_TREE, OnKeydownTree)
	ON_NOTIFY(TVN_BEGINDRAG, IDC_TREE, OnBegindragTree)
	ON_WM_LBUTTONUP()
	ON_WM_RBUTTONDOWN()
	ON_WM_TIMER()
	ON_WM_SIZE()
	//}}AFX_MSG_MAP
END_MESSAGE_MAP()

/////////////////////////////////////////////////////////////////////////////
// CDragDropView diagnostics

#ifdef _DEBUG
void CDragDropView::AssertValid() const
{
	CFormView::AssertValid();
}

void CDragDropView::Dump(CDumpContext& dc) const
{
	CFormView::Dump(dc);
}
#endif //_DEBUG

/////////////////////////////////////////////////////////////////////////////
// CDragDropView message handlers

void CDragDropView::OnInitialUpdate() 
{
	CFormView::OnInitialUpdate();
	
  m_Tree.SetImageList(&m_imgList,TVSIL_NORMAL);

  // now that the tree exists, force resize

  RECT rect;
  GetWindowRect(&rect);
  OnSize(SIZE_RESTORED,rect.right-rect.left,rect.bottom-rect.top);

  // populate tree

  HTREEITEM hScene  = m_Tree.InsertItem("scene",     1,1);
  HTREEITEM hCams   = m_Tree.InsertItem("cameras",   1,1,hScene);
  HTREEITEM hLights = m_Tree.InsertItem("lights",    1,1,hScene);
  HTREEITEM hMeshes = m_Tree.InsertItem("meshes",    9,9,hScene);
  HTREEITEM hChrs   = m_Tree.InsertItem("characters",1,1,hMeshes);
  HTREEITEM hBlocks = m_Tree.InsertItem("blocks",    1,1,hMeshes);
  HTREEITEM hHouses = m_Tree.InsertItem("buildings", 1,1,hMeshes);
  CString strTmp;

  for(int i = 0; i < 10; i++)
  {
    strTmp.Format("camera %u",i+1);
    m_Tree.InsertItem(LPCTSTR(strTmp),3,3,hCams);
    strTmp.Format("light %u",i+1);
    m_Tree.InsertItem(LPCTSTR(strTmp),6,6,hLights);
    strTmp.Format("character %u",i+1);
    m_Tree.InsertItem(LPCTSTR(strTmp),13,13,hChrs);
    strTmp.Format("block %u",i+1);
    m_Tree.InsertItem(LPCTSTR(strTmp),10,10,hBlocks);
    strTmp.Format("house %u",i+1);
    m_Tree.InsertItem(LPCTSTR(strTmp),11,11,hHouses);
  }

//  m_Tree.Expand(hHouses,TVE_EXPAND);
//  m_Tree.Expand(hBlocks,TVE_EXPAND);
  m_Tree.Expand(hChrs,  TVE_EXPAND);
  m_Tree.Expand(hMeshes,TVE_EXPAND);
  m_Tree.Expand(hLights,TVE_EXPAND);
  m_Tree.Expand(hCams,  TVE_EXPAND);
  m_Tree.Expand(hScene, TVE_EXPAND);
}

//

void CDragDropView::OnSize(UINT nType, int cx, int cy) 
{
	CFormView::OnSize(nType, cx, cy);
	
  // resize tree

  if(m_Tree.m_hWnd)
  {                                                         // otherwise not created yet
    RECT rect;
    GetClientRect(&rect);
    #define BORDER 5
    int iSizeX,iSizeY;
    iSizeX = rect.right-rect.left-BORDER*2;
    iSizeY = rect.bottom-rect.top-BORDER*2;
    if((iSizeX > 0) && (iSizeY > 0)) m_Tree.SetWindowPos(NULL,BORDER,BORDER,iSizeX,iSizeY,SWP_NOZORDER);
    #undef BORDER
  }
}

//

void CDragDropView::OnKeydownTree(NMHDR* pNMHDR, LRESULT* pResult) 
{
	TV_KEYDOWN*  pTVKeyDown = (TV_KEYDOWN*)pNMHDR;
  WORD wVKey = pTVKeyDown->wVKey;

  if(wVKey == VK_CONTROL) DragToggleItem();
	
	*pResult = 0;
}

// collapse/expand item

void CDragDropView::DragToggleItem()
{
  if(m_boDragging) 
  {
    if(m_hDragTarget)
    {
      m_pDragImgList->DragShowNolock(false);
      m_Tree.Expand(m_hDragTarget,TVE_TOGGLE);
      m_Tree.RedrawWindow();
      m_pDragImgList->DragShowNolock(true);
    }
  }
}

// begin drag

void CDragDropView::OnBegindragTree(NMHDR* pNMHDR, LRESULT* pResult) 
{
	NM_TREEVIEW* pNMTreeView = (NM_TREEVIEW*)pNMHDR;
  HTREEITEM    hItem       = pNMTreeView->itemNew.hItem;
  if(!hItem) return;
  m_Tree.SelectItem(hItem);

  m_hDragItem    = hItem;
  m_pDragImgList = m_Tree.CreateDragImage(hItem);
  if(!m_pDragImgList) return;
  
  m_pDragImgList->BeginDrag(0,CPoint(0,0));
  m_pDragImgList->DragEnter(this,pNMTreeView->ptDrag);
  m_Tree.m_pDragImgList = m_pDragImgList;                   // for CTreeCtrlDrag::OnMouseWheel
  m_boDragging = true;

  ShowCursor(false);
  SetCapture();
  m_hDragTarget = NULL;

  m_nTimer = SetTimer(1,25,NULL);
	*pResult = 0;
}

// end drag

void CDragDropView::OnLButtonUp(UINT nFlags, CPoint point) 
{
  if(m_boDragging)
  {
    KillTimer(m_nTimer);

    ReleaseCapture();
    ShowCursor(true);
    m_boDragging = false;

    m_pDragImgList->DragLeave(this);
    m_pDragImgList->EndDrag();
    m_Tree.m_pDragImgList = NULL;                           // CTreeCtrlDrag
    delete m_pDragImgList;
    m_pDragImgList = NULL;

    m_Tree.SelectDropTarget(NULL);
    m_Tree.SelectItem(m_hDragItem);

    if(m_hDragTarget && (m_hDragTarget != m_hDragItem))
    {
      SuccessfulDrag(m_hDragTarget,m_hDragItem);
    }
  } 	
	
	CFormView::OnLButtonUp(nFlags, point);
}

// abort drag

void CDragDropView::OnRButtonDown(UINT nFlags, CPoint point) 
{
  if(m_boDragging)
  {
    KillTimer(m_nTimer);

    ReleaseCapture();
    ShowCursor(true);
    m_boDragging = false;

    m_pDragImgList->DragLeave(this);
    m_pDragImgList->EndDrag();
    m_Tree.m_pDragImgList = NULL;                           // CTreeCtrlDrag
    delete m_pDragImgList;
    m_pDragImgList = NULL;

    m_Tree.SelectDropTarget(NULL);
    m_Tree.SelectItem(m_hDragItem);
    m_Tree.EnsureVisible(m_hDragItem);
  } 	
	
	CFormView::OnRButtonDown(nFlags, point);
}

// highlight & scroll

void CDragDropView::OnTimer(UINT nIDEvent) 
{
  if(m_boDragging)
  {
    POINT point;
    GetCursorPos(&point);
    ScreenToClient(&point);

    // highlight target

    TVHITTESTINFO tvHit;
    tvHit.pt = point;
    HTREEITEM hTarget = m_Tree.HitTest(&tvHit);

    if(hTarget)
    {
      if(hTarget != m_hDragTarget)
      {                                                     // this test avoids flickering
        m_pDragImgList->DragShowNolock(false);
        m_Tree.EnsureVisible(hTarget);
        m_Tree.SelectDropTarget(hTarget);
        m_pDragImgList->DragShowNolock(true);
        m_hDragTarget = hTarget;
      }
    }

    // scroll tree

    else
    {
      RECT rect;
      m_Tree.GetClientRect(&rect);

      int iMaxV = m_Tree.GetScrollLimit(SB_VERT);
 	    int iPosV = m_Tree.GetScrollPos  (SB_VERT);

      // up
      if((point.y < rect.top -10) && iPosV)
      {
        HTREEITEM hPrev = m_Tree.GetPrevVisibleItem(m_Tree.GetFirstVisibleItem());
        m_pDragImgList->DragShowNolock(false);
        m_Tree.EnsureVisible(hPrev);
        m_pDragImgList->DragShowNolock(true);
      }

      // down
      if((point.y > rect.bottom +10) && (iPosV != iMaxV))
      {
        UINT Nb = m_Tree.GetVisibleCount();
        if(Nb != -1)
        {
          HTREEITEM hNext = m_Tree.GetFirstVisibleItem();
          for(UINT i = 0; i < Nb; i++) hNext = m_Tree.GetNextVisibleItem(hNext);
          m_pDragImgList->DragShowNolock(false);
          m_Tree.EnsureVisible(hNext);
          m_pDragImgList->DragShowNolock(true);
        }
      }

      int iPosH = m_Tree.GetScrollPos  (SB_HORZ);
    	int iMaxH = m_Tree.GetScrollLimit(SB_HORZ);

      // left
      if((point.x < rect.left) && iPosH)
      {
        m_pDragImgList->DragShowNolock(false);
        m_Tree.SendMessage(WM_HSCROLL,SB_LINELEFT);
        m_pDragImgList->DragShowNolock(true);
      }

      // right
      if((point.x > rect.right) && (iPosH != iMaxH))
      {
        m_pDragImgList->DragShowNolock(false);
        m_Tree.SendMessage(WM_HSCROLL,SB_LINERIGHT);
        m_pDragImgList->DragShowNolock(true);
      }
    }

    m_pDragImgList->DragMove(point);
  }
	
	CFormView::OnTimer(nIDEvent);
}

//
// result of the drag & drop operation

void CDragDropView::SuccessfulDrag(HTREEITEM hDest,HTREEITEM hSrc)
{
  // test : no op if hSrc is already a child of hDest

  if(m_Tree.GetParentItem(hSrc) == hDest) return;

  // test : if hDest is a child of hSrc, we're going to have a stack overflow ! (infinite loop)

  HTREEITEM hParent = hDest;
  while(hParent)
  {
    if(hParent == hSrc)
    {
      MessageBox("invalid drag & drop : destination can't be a child of source (infinite loop)","cancelled",MB_ICONEXCLAMATION);
      return;
    }
    hParent = m_Tree.GetParentItem(hParent);
  }

  // create a copy of the source subtree

  HTREEITEM hNew = m_Tree.InsertItemAndSubtree(hDest,hSrc);
  if(!hNew) return;

  CString strTmp("drag & drop : ");
  strTmp += "[" + m_Tree.GetItemText(hSrc) + "] becomes a child of [" + m_Tree.GetItemText(hDest) +"]";
  //MessageBox((LPCTSTR)strTmp,"drag & drop",MB_OK);

  //

  m_Tree.SelectItem   (hNew);
  m_Tree.EnsureVisible(hNew);
  m_Tree.DeleteItem   (hSrc);
}
