// DragDropView.cpp : implementation of the CDragDropView class
//

#include "stdafx.h"
#include "DragDrop.h"

#include "DragDropDoc.h"
#include "DragDropView.h"

#ifdef _DEBUG
#define new DEBUG_NEW
#undef THIS_FILE
static char THIS_FILE[] = __FILE__;
#endif

/////////////////////////////////////////////////////////////////////////////
// CDragDropView

IMPLEMENT_DYNCREATE(CDragDropView, CTreeView)

BEGIN_MESSAGE_MAP(CDragDropView, CTreeView)
	//{{AFX_MSG_MAP(CDragDropView)
	ON_NOTIFY_REFLECT(TVN_KEYDOWN, OnKeydown)
	ON_NOTIFY_REFLECT(TVN_BEGINDRAG, OnBegindrag)
	ON_WM_LBUTTONUP()
	ON_WM_RBUTTONDOWN()
	ON_WM_TIMER()
	ON_WM_MOUSEWHEEL()
	//}}AFX_MSG_MAP
END_MESSAGE_MAP()

/////////////////////////////////////////////////////////////////////////////
// CDragDropView construction/destruction

CDragDropView::CDragDropView()
{
  m_boDragging   = false;
  m_pDragImgList = NULL;
	m_imgList.Create(IDB_SCENE_TREE,16,1,RGB(255,255,255));
}

//

CDragDropView::~CDragDropView()
{
  delete m_pDragImgList;
  m_imgList.DeleteImageList();
}

//

BOOL CDragDropView::PreCreateWindow(CREATESTRUCT& cs)
{
	BOOL Res  = CTreeView::PreCreateWindow(cs);
  cs.style |= TVS_FULLROWSELECT | TVS_HASBUTTONS | TVS_HASLINES | TVS_LINESATROOT | TVS_SHOWSELALWAYS;
  return Res;
}

/////////////////////////////////////////////////////////////////////////////
// CDragDropView drawing

void CDragDropView::OnDraw(CDC* pDC)
{
	CDragDropDoc* pDoc = GetDocument();
	ASSERT_VALID(pDoc);
	// TODO: add draw code for native data here
}

//

void CDragDropView::OnInitialUpdate()
{
	CTreeView::OnInitialUpdate();

  CTreeCtrl& Tree = GetTreeCtrl();
  Tree.SetImageList(&m_imgList,TVSIL_NORMAL);

  // populate tree

  HTREEITEM hScene  = Tree.InsertItem("scene",     1,1);
  HTREEITEM hCams   = Tree.InsertItem("cameras",   1,1,hScene);
  HTREEITEM hLights = Tree.InsertItem("lights",    1,1,hScene);
  HTREEITEM hMeshes = Tree.InsertItem("meshes",    9,9,hScene);
  HTREEITEM hChrs   = Tree.InsertItem("characters",1,1,hMeshes);
  HTREEITEM hBlocks = Tree.InsertItem("blocks",    1,1,hMeshes);
  HTREEITEM hHouses = Tree.InsertItem("buildings", 1,1,hMeshes);
  CString strTmp;

  for(int i = 0; i < 10; i++)
  {
    strTmp.Format("camera %u",i+1);
    Tree.InsertItem(LPCTSTR(strTmp),3,3,hCams);
    strTmp.Format("light %u",i+1);
    Tree.InsertItem(LPCTSTR(strTmp),6,6,hLights);
    strTmp.Format("character %u",i+1);
    Tree.InsertItem(LPCTSTR(strTmp),13,13,hChrs);
    strTmp.Format("block %u",i+1);
    Tree.InsertItem(LPCTSTR(strTmp),10,10,hBlocks);
    strTmp.Format("house %u",i+1);
    Tree.InsertItem(LPCTSTR(strTmp),11,11,hHouses);
  }

//  Tree.Expand(hHouses,TVE_EXPAND);
//  Tree.Expand(hBlocks,TVE_EXPAND);
  Tree.Expand(hChrs,  TVE_EXPAND);
  Tree.Expand(hMeshes,TVE_EXPAND);
  Tree.Expand(hLights,TVE_EXPAND);
  Tree.Expand(hCams,  TVE_EXPAND);
  Tree.Expand(hScene, TVE_EXPAND);
}

/////////////////////////////////////////////////////////////////////////////
// CDragDropView diagnostics

#ifdef _DEBUG
void CDragDropView::AssertValid() const
{
	CTreeView::AssertValid();
}

void CDragDropView::Dump(CDumpContext& dc) const
{
	CTreeView::Dump(dc);
}

CDragDropDoc* CDragDropView::GetDocument() // non-debug version is inline
{
	ASSERT(m_pDocument->IsKindOf(RUNTIME_CLASS(CDragDropDoc)));
	return (CDragDropDoc*)m_pDocument;
}
#endif //_DEBUG

/////////////////////////////////////////////////////////////////////////////
// CDragDropView message handlers

void CDragDropView::OnKeydown(NMHDR* pNMHDR, LRESULT* pResult) 
{
	TV_KEYDOWN*  pTVKeyDown = (TV_KEYDOWN*)pNMHDR;
  WORD wVKey = pTVKeyDown->wVKey;

  if(wVKey == VK_CONTROL) DragToggleItem();
	
	*pResult = 0;
}

// collapse/expand item

void CDragDropView::DragToggleItem()
{
  if(m_boDragging) 
  {
    if(m_hDragTarget)
    {
      m_pDragImgList->DragShowNolock(false);
      GetTreeCtrl().Expand(m_hDragTarget,TVE_TOGGLE);
      GetTreeCtrl().RedrawWindow();
      m_pDragImgList->DragShowNolock(true);
    }
  }
}

// begin drag

void CDragDropView::OnBegindrag(NMHDR* pNMHDR, LRESULT* pResult) 
{
	NM_TREEVIEW* pNMTreeView = (NM_TREEVIEW*)pNMHDR;
  HTREEITEM    hItem       = pNMTreeView->itemNew.hItem;
  if(!hItem) return;
  GetTreeCtrl().SelectItem(hItem);

  m_hDragItem    = hItem;
  m_pDragImgList = GetTreeCtrl().CreateDragImage(hItem);
  if(!m_pDragImgList) return;
  
  m_pDragImgList->BeginDrag(0,CPoint(0,0));
  m_pDragImgList->DragEnter(this,pNMTreeView->ptDrag);
  m_boDragging = true;

  ShowCursor(false);
  SetCapture();
  m_hDragTarget = NULL;

  m_nTimer = SetTimer(1,25,NULL);
	*pResult = 0;
}

// end drag

void CDragDropView::OnLButtonUp(UINT nFlags, CPoint point) 
{
  if(m_boDragging)
  {
    KillTimer(m_nTimer);

    ReleaseCapture();
    ShowCursor(true);
    m_boDragging = false;

    m_pDragImgList->DragLeave(this);
    m_pDragImgList->EndDrag();
    delete m_pDragImgList;
    m_pDragImgList = NULL;

    GetTreeCtrl().SelectDropTarget(NULL);
    GetTreeCtrl().SelectItem(m_hDragItem);

    if(m_hDragTarget && (m_hDragTarget != m_hDragItem))
    {
      SuccessfulDrag(m_hDragTarget,m_hDragItem);
    }
  } 	
	
	CTreeView::OnLButtonUp(nFlags, point);
}

// abort drag

void CDragDropView::OnRButtonDown(UINT nFlags, CPoint point) 
{
  if(m_boDragging)
  {
    KillTimer(m_nTimer);

    ReleaseCapture();
    ShowCursor(true);
    m_boDragging = false;

    m_pDragImgList->DragLeave(this);
    m_pDragImgList->EndDrag();
    delete m_pDragImgList;
    m_pDragImgList = NULL;

    GetTreeCtrl().SelectDropTarget(NULL);
    GetTreeCtrl().SelectItem(m_hDragItem);
    GetTreeCtrl().EnsureVisible(m_hDragItem);
  } 	
	
	CTreeView::OnRButtonDown(nFlags, point);
}

// highlight & scroll

void CDragDropView::OnTimer(UINT nIDEvent)
{
  if(m_boDragging)
  {
    POINT point;
    GetCursorPos(&point);
    ScreenToClient(&point);

    // highlight target

    CTreeCtrl& Tree = GetTreeCtrl();
    TVHITTESTINFO tvHit;
    tvHit.pt = point;
    HTREEITEM hTarget = Tree.HitTest(&tvHit);

    if(hTarget)
    {
      if(hTarget != m_hDragTarget)
      {                                                     // this test avoids flickering
        m_pDragImgList->DragShowNolock(false);
        Tree.EnsureVisible(hTarget);
        Tree.SelectDropTarget(hTarget);
        m_pDragImgList->DragShowNolock(true);
        m_hDragTarget = hTarget;
      }
    }

    // scroll tree

    else
    {
      RECT rect;
      Tree.GetClientRect(&rect);

      int iMaxV = Tree.GetScrollLimit(SB_VERT);
 	    int iPosV = Tree.GetScrollPos  (SB_VERT);

      // up
      if((point.y < rect.top -10) && iPosV)
      {
        HTREEITEM hPrev = Tree.GetPrevVisibleItem(Tree.GetFirstVisibleItem());
        m_pDragImgList->DragShowNolock(false);
        Tree.EnsureVisible(hPrev);
        m_pDragImgList->DragShowNolock(true);
      }

      // down
      if((point.y > rect.bottom +10) && (iPosV != iMaxV))
      {
        UINT Nb = Tree.GetVisibleCount();
        if(Nb != -1)
        {
          HTREEITEM hNext = Tree.GetFirstVisibleItem();
          for(UINT i = 0; i < Nb; i++) hNext = Tree.GetNextVisibleItem(hNext);
          m_pDragImgList->DragShowNolock(false);
          Tree.EnsureVisible(hNext);
          m_pDragImgList->DragShowNolock(true);
        }
      }

      int iPosH = Tree.GetScrollPos  (SB_HORZ);
    	int iMaxH = Tree.GetScrollLimit(SB_HORZ);

      // left
      if((point.x < rect.left) && iPosH)
      {
        m_pDragImgList->DragShowNolock(false);
        Tree.SendMessage(WM_HSCROLL,SB_LINELEFT);
        m_pDragImgList->DragShowNolock(true);
      }

      // right
      if((point.x > rect.right) && (iPosH != iMaxH))
      {
        m_pDragImgList->DragShowNolock(false);
        Tree.SendMessage(WM_HSCROLL,SB_LINERIGHT);
        m_pDragImgList->DragShowNolock(true);
      }
    }

    m_pDragImgList->DragMove(point);
  }
	
	CTreeView::OnTimer(nIDEvent);
}

//

BOOL CDragDropView::OnMouseWheel(UINT nFlags, short zDelta, CPoint pt) 
{
  if(m_pDragImgList) m_pDragImgList->DragShowNolock(false);
  BOOL Res = CTreeView::OnMouseWheel(nFlags, zDelta, pt);
  if(m_pDragImgList) m_pDragImgList->DragShowNolock(true);
  return Res;
}

//
// result of the drag & drop operation

void CDragDropView::SuccessfulDrag(HTREEITEM hDest,HTREEITEM hSrc)
{
  // test : no op if hSrc is already a child of hDest

  CTreeCtrl& Tree = GetTreeCtrl();
  if(Tree.GetParentItem(hSrc) == hDest) return;

  // test : if hDest is a child of hSrc, we're going to have a stack overflow ! (infinite loop)

  HTREEITEM hParent = hDest;
  while(hParent)
  {
    if(hParent == hSrc)
    {
      MessageBox("invalid drag & drop : destination can't be a child of source (infinite loop)","cancelled",MB_ICONEXCLAMATION);
      return;
    }
    hParent = Tree.GetParentItem(hParent);
  }

  // create a copy of the source subtree

  HTREEITEM hNew = InsertItemAndSubtree(hDest,hSrc);
  if(!hNew) return;

  CString strTmp("drag & drop : ");
  strTmp += "[" + Tree.GetItemText(hSrc) + "] becomes a child of [" + Tree.GetItemText(hDest) +"]";
  //MessageBox((LPCTSTR)strTmp,"drag & drop",MB_OK);

  //

  Tree.SelectItem   (hNew);
  Tree.EnsureVisible(hNew);
  Tree.DeleteItem   (hSrc);
}

//

HTREEITEM CDragDropView::InsertItemAndSubtree(HTREEITEM hDest,HTREEITEM hSrc)
{
  if(!hDest || !hSrc) return NULL;

  HTREEITEM hNew = GetTreeCtrl().InsertItem("new",hDest);
  if(!hNew) return NULL;

  CopySubtree(hNew,hSrc);                                   // start with subtree so that item is correctly expanded
  CopyItem   (hNew,hSrc);
  return hNew;
}

//

void CDragDropView::CopySubtree(HTREEITEM hDest,HTREEITEM hSrc)
{
  if(!hDest || !hSrc) return;
  CTreeCtrl& Tree = GetTreeCtrl();

  HTREEITEM hChildSrc = Tree.GetChildItem(hSrc);
  while(hChildSrc)
  {
    HTREEITEM   hChildDest = Tree.InsertItem("dest child",hDest);
    CopySubtree(hChildDest,hChildSrc);
    CopyItem   (hChildDest,hChildSrc);

    hChildSrc = Tree.GetNextSiblingItem(hChildSrc);
  }
}

//

void CDragDropView::CopyItem(HTREEITEM hDest,HTREEITEM hSrc)
{
  CTreeCtrl& Tree = GetTreeCtrl();

  TV_ITEM tvSrc;
  tvSrc.mask  = TVIF_HANDLE | TVIF_PARAM | TVIF_IMAGE | TVIF_SELECTEDIMAGE | TVIF_STATE;
  tvSrc.hItem = hSrc;
  if(!Tree.GetItem(&tvSrc)) return;

  tvSrc.hItem = hDest;
  Tree.SetItem(&tvSrc);
  Tree.SetItemText(hDest,Tree.GetItemText(hSrc));
  Tree.SetCheck   (hDest,Tree.GetCheck   (hSrc));

  if(tvSrc.state & TVIS_EXPANDED) Tree.Expand(hDest,TVE_EXPAND);
}
